export PYTEST_MARKER="-m cassandra"
export DOCKER_DEPS="cassandra3"
export CASSANDRA_HOST="cassandra3"
export WAIT_FOR_HOST="cassandra3"
export WAIT_FOR_PORT=9042
export CASS_DRIVER_BUILD_CONCURRENCY=4
