#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule

logger = logging.getLogger("elasticapm")


class DjangoAutoInstrumentation(AbstractInstrumentedModule):
    name = "django"

    instrument_list = [("django.apps.registry", "Apps.populate")]

    creates_transactions = True

    def call(self, module, method, wrapped, instance, args, kwargs) -> None:
        installed_apps = list(args[0] if args else kwargs["installed_apps"])
        if (
            "elasticapm.contrib.django" not in installed_apps
            and "elasticapm.contrib.django.apps.ElasticAPMConfig" not in installed_apps
        ):
            logger.info("Prepending elasticapm to Django INSTALLED_APPS")
            installed_apps.insert(0, "elasticapm.contrib.django")
            if args:
                args = list(args)
                args[0] = installed_apps
            else:
                kwargs["installed_apps"] = installed_apps
        wrapped(*args, **kwargs)
