export PYTEST_MARKER="-m aiomysql"
export DOCKER_DEPS="mysql"
export MYSQL_HOST="mysql"
export MYSQL_USER="eapm"
export MYSQL_PASSWORD="Very(!)Secure"
export WAIT_FOR_HOST="mysql"
export WAIT_FOR_PORT=3306
