export PYTEST_MARKER="-m psycopg_pool"
export DOCKER_DEPS="postgres"
export POSTGRES_HOST="postgres"
export POSTGRES_USER="postgres"
export POSTGRES_PASSWORD="postgres"
export POSTGRES_DB="elasticapm_test"
export POSTGRES_PORT="5432"
